"""Serializing data to text."""

import string
import base64
import math
import datetime
import vobject
import parsing

from dateutil import relativedelta
from dateutil import rrule
from dateutil import parser
from dateutil import tz

from datastructures import *

from vobject import ContentLine




#todo:
# with datetimes, serious timezone and utc features. possibly subclass python's datetime.datetime?



## Utils

def numToDigits(num, places):
    """Helper, for converting numbers to textual digits."""
    s = str(num)
    if len(s) < places:
        return ("0" * (places - len(s))) + s
    elif len(s) > places:
        return s[len(s)-places: ]
    else:
        return s

def floor(x, y):
    i = math.floor(x/y)
    rem = x % y
    return i,rem


## Serializing values of contentlines

#  see below for serializing entire contentlines

def calAddressToString(addr):
    """Returns a URI (stored in a string)."""
    return uriToString(addr)

def uriToString(uri):
    """Returns a string URI."""
    return uri

def booleanToString(b):
    if b: return "TRUE"
    else: return "FALSE"

def binaryToString(b):
    return base64.encodestring(b)

def floatToString(num):
    return str(num)

def integerToString(num):
    return str(num)

def timedeltaToString(delta):
    """Convert timedelta to an rfc2445 DURATION."""
    if delta.days == 0: sign = 1
    else: sign = delta.days / abs(delta.days)
    delta = abs(delta)
    days = delta.days
    hours = delta.seconds / 3600
    minutes = (delta.seconds % 3600) / 60
    seconds = delta.seconds % 60
    out = ''
    if sign == -1: out = '-'
    out += 'P'
    if days: out += str(days) + 'D'
    if hours or minutes or seconds: out += 'T'
    elif not days: #Deal with zero duration
        out += '0S'
    if hours: out += str(hours) + 'H'
    if minutes: out += str(minutes) + 'M'
    if seconds: out += str(seconds) + 'S'
    return out
    

def durationsToString(durationsInTimeDelta):
    days = durationsInTimeDelta.days
    secs = durationsInTimeDelta.seconds
    daysString = str(days) + "H"
    secsString = str(secs) + "S"
    return daysString + secsString

def testNumToDigits():
    assert( numToDigits(2, 3)       == '002' )
    assert( numToDigits(000002, 3)  == '002' )
    assert( numToDigits(2000002, 3) == '002' )
    assert( numToDigits(202, 3)     == '202' )

def dateToString(date):
    year  = numToDigits( date.year,  4 )
    month = numToDigits( date.month, 2 )
    day   = numToDigits( date.day,   2 )
    return year + month + day

def dateTimeToString(dateTime, preserveTZ=False):
    """Convert to UTC if tzinfo is set, unless preserveTZ.  Output string."""
    if dateTime.tzinfo and not preserveTZ:
        dateTime = dateTime.astimezone(tz.tzutc())
    if dateTime.tzinfo == tz.tzutc(): utcString = "Z"
    else: utcString = ""

    year  = numToDigits( dateTime.year,  4 )
    month = numToDigits( dateTime.month, 2 )
    day   = numToDigits( dateTime.day,   2 )
    hour  = numToDigits( dateTime.hour,  2 )
    mins  = numToDigits( dateTime.minute,  2 )
    secs  = numToDigits( dateTime.second,  2 )

    return year + month + day + "T" + hour + mins + secs + utcString

def geoToString (geoPosition):
    return str(geoPosition.latitude) + ";" + str(geoPosition.longitude) 

def durationTimeDeltaToString(duration):
    """Convert datastructures.Duration object to a string."""
    sign = duration.sign
    timedelta = duration.timedelta
    info = {week:timedelta.week, day:timedelta.day, hour:timedelta.hour, minute:timedelta.minute, second:timedelta.second}
    code = {week:"W", day:"D", hour:"H", minute:"M", second:"S"}
    result = ""
    if sign == -1:
        result += "-"
    result += "P"
    for key in info.keys():
        val = info[key]
        if not val == 0:
            result += ( str(val) + code[key] )
    return result

def tzOffsetToString (tzoff):
    name   = tzoff.name
    offset = tzoff.offset
    sign = False
    if offset < 0:
        offset = abs(offset)
        sign = True
    hours,remSecs    = floor(offset, 60*60)
    minutes, seconds = floor(remSecs, 60)
    hoursString      = numToDigits(hours, 2)
    minutesString    = numToDigits(minutes, 2)
    secondsString    = ""     
    signString       = ""
    if seconds != 0:
        secondsString = numToDigits(seconds, 2)   
    if sign:
        signString = "-"
    return sign + hoursString + minutesString + secondsString

def textToString(s):
    result = ""
    for c in s:
        if c in parsing.escapableCharList:
            result += "\\" + c
        else: 
            result += c
    return result

def textValuesToString(strings):
    results = []
    for s in strings:
        results.append(textToString(s))
    return string.join(results, ",")

def recurrenceToString(recur):
    return recur.string



## Serializing contentlines

def serializeAttendee(contentlineName, attendee):
    address = attendee.calAddress
    table   = attendee.fields
    return ContentLine(contentlineName, table, address)

def serializeOrganizer(contentlineName, organizer):
    address = organizer.calAddress
    table   = organizer.fields
    return ContentLine(contentlineName, table, address)

def serializeFreeBusy(contentlineName, freebusy):
    periods = freebusy.periods
    type    = freebusy.type
    table   = {"FBTYPE" : type}
    return ContentLine( contentlineName, table, string.join(periods, ",") )

def serializeBeginEnd(contentlineName, value):
    return ContentLine(contentlineName, {}, value)

def serializeAttachment (contentlineName, attachment):
    if isinstance(attachment, str):
        # encodestring does this shockingly retarded thing where always adds a newline
        return ContentLine(contentlineName,
                           {"VALUE":"BINARY", "ENCODING":"BASE64"},
                           base64.encodestring(attachment)[ :-1])
    else:
        # attachment is of type Attachment
        return ContentLine( contentlineName,
                            {"FMTTYPE" : attachment.format},
                            uriToString(attachment.url) )

def serializeInteger(contentlineName, value):
    return ContentLine( contentlineName, {}, integerToString(value) )

def serializeGeo(contentlineName, value):
    return ContentLine( contentlineName, {}, geoToString(value) )

def serializeDateTime(contentlineName, value):
    return ContentLine( contentlineName, {}, dateTimeToString(value) )

def serializeURI(contentlineName, value):
    return ContentLine( contentlineName, {}, uriToString(value) )

def serializeRdate (contentlineName, value):
    if isinstance(value, DateRdate):
        table = {"VALUE":"DATE"}
    elif isinstance(value, PeriodRdate):
        table = {"VALUE":"PERIOD"}
    else:
        table = {"VALUE":"DATE-TIME"}
    if value.tzid != None:
        table["TZID"] = value.tzid
    return ContentLine(contentlineName, table, string.join(value.values, ","))

def serializeText(contentlineName, value):
    table = {"ALTREP":value.altRepresentation, "LANGUAGE":value.language}
    return ContentLine( contentlineName, table, textValuesToString(value.values) )

def serializeText(contentlineName, value):
    return ContentLine( contentlineName, {}, tzOffsetToString(value) )

def serializeDtstart (contentlineName, value):
    if isinstance(value, datetime.date):
        table = {"VALUE":"DATE"}
        return ContentLine( contentlineName, table, dateToString(value) )
    else:
        table = {"VALUE":"DATE-TIME"}
        return ContentLine( contentlineName, table, dateTimeToString(value) )

def serializeDuration (contentlineName, value):
    if isinstance(value, datetime.datetime):
        return ContentLine( contentlineName, {}, dateTimeToString(value) )
    else:
        return ContentLine( contentlineName, {}, durationTimeDeltaToString(value) )

def serializeRecur (contentlineName, value):
    return ContentLine( contentlineName, {}, recurrenceToString(value) )

def serializeTrigger (contentlineName, value):
    if isinstance(value, datetime.date):
        table = {"VALUE":"DATE"}
        return ContentLine( contentlineName, table, dateToString(value) )
    elif isinstance(value, datetime.datetime):
        table = {"VALUE":"DATE-TIME"}
        return ContentLine( contentlineName, table, dateTimeToString(value) )
    else:
        table = {"RELATED":value.related}
        return ContentLine( contentlineName, table, durationTimeDeltaToString(value.duration) )


def test():
    # serializeDateTime
    print "serializeDateTime:"
    dtstring = "COMPLETED:19970630T235959Z"
    contentline = vobject.textLineToContentLine(dtstring)
    dtparse  = parsing.parse(contentline)
    print dateTimeToString(dtparse) == string.split(dtstring, ":")[1]
    print contentline
    print dtparse.__repr__()
    print serializeDateTime("COMPLETED", dtparse)
    print serializeDateTime("COMPLETED", dtparse) == contentline #ContentLine equality test


    

#------------------- Testing and running functions -----------------------------
if __name__ == '__main__':
    test()
